from typing import Any, Dict

from lowerpines.gmi import GMI
from lowerpines.endpoints.request import Request


class SmsCreateRequest(Request):
    def __init__(self, gmi: GMI, duration: int, registration_id: str) -> None:
        if duration > 48:
            raise ValueError('Cannot have a duration of SMS mode for more than 48 hours')
        self.duration = duration
        self.registration_id = registration_id
        super().__init__(gmi)

    def mode(self) -> str:
        return "POST"

    def parse(self, response: dict) -> Any:
        return None

    def args(self) -> Dict[str, Any]:
        return {
            'duration': self.duration,
            'registration_id': self.registration_id
        }

    def url(self) -> str:
        return self.base_url + "/users/sms_mode"


class SmsDeleteRequest(Request):
    def mode(self) -> str:
        return "POST"

    def parse(self, response: dict) -> Any:
        return None

    def url(self) -> str:
        return self.base_url + '/users/sms_mode/delete'



from lowerpines.exceptions import NoneFoundException, MultipleFoundException


class AbstractManager:
    def __len__(self):
        self.lazy_fill_content()
        return self._content.__len__()

    def __getitem__(self, key):
        self.lazy_fill_content()
        return self._content.__getitem__(key)

    def __iter__(self):
        self.lazy_fill_content()
        return self._content.__iter__()

    def __init__(self, gmi, content=None):
        self.gmi = gmi
        self._content = content

    def _all(self):
        raise NotImplementedError

    def get(self, **kwargs):
        filtered = self.filter(**kwargs)
        if len(filtered) == 0:
            raise NoneFoundException()
        elif len(filtered) == 1:
            return filtered[0]
        else:
            raise MultipleFoundException()

    def filter(self, **kwargs):
        self.lazy_fill_content()
        filtered = self._content
        for arg, value in kwargs.items():
            filtered = [item for item in filtered if getattr(item, arg) == value]
        return self.__class__(self.gmi, filtered)

    def lazy_fill_content(self):
        if self._content is None:
            self._content = self._all()



from lowerpines.endpoints.object import AbstractObject, Field, RetrievableObject
from lowerpines.endpoints.request import Request
from lowerpines.endpoints.like import LikeCreateRequest, LikeDestroyRequest
from lowerpines.exceptions import InvalidOperationException


class Message(AbstractObject, RetrievableObject):
    message_id = Field(api_name='id')
    source_guid = Field()
    created_at = Field()
    user_id = Field()
    group_id = Field()
    name = Field()
    avatar_url = Field()
    text = Field()
    system = Field()
    favorited_by = Field()
    attachments = Field(handler=None)
    sender_type = Field()
    sender_id = Field()

    def __init__(self, gmi, group_id=None, source_guid=None, text=None, attachments=list()):
        self.gmi = gmi
        self.group_id = group_id
        self.source_guid = source_guid
        self.text = text
        self.attachments = attachments
        self.message_id = None

    def save(self):
        if self.message_id:
            raise InvalidOperationException("You cannot change a message that has already been sent")
        else:
            new_data = MessagesCreateRequest(self.gmi, self.group_id, self.source_guid, self.text,
                                             self.attachments).result
            self._refresh_from_other(new_data)

    def refresh(self):
        if self.message_id:
            new_data = MessagesShowRequest(self.gmi, self.group_id, self.message_id).result
            self._refresh_from_other(new_data)
        else:
            raise InvalidOperationException("Must have a message_id to pull data from the server")

    def on_fields_loaded(self):
        if self.text is None:
            self.text = ""
        from lowerpines.message import ComplexMessage, RefAttach
        self.complex_text = ComplexMessage('')
        doing_mentions = False
        for attachment in self.attachments:
            if attachment['type'] == 'mentions':
                doing_mentions = True
                prev_index = 0
                for i in range(len(self.text)):
                    for loci, user_id in zip(attachment['loci'], attachment['user_ids']):
                        if loci[0] == i:
                            self.complex_text += self.text[prev_index:loci[0]] + \
                                                    RefAttach(user_id, self.text[loci[0]:loci[0] + loci[1]])
                            prev_index = loci[0] + loci[1]
                self.complex_text = self.complex_text + self.text[prev_index:]
        if not doing_mentions:
            self.complex_text = ComplexMessage(self.text)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.text

    @staticmethod
    def get(gmi, group_id, message_id):
        return MessagesShowRequest(gmi, group_id, message_id).result

    def like(self):
        LikeCreateRequest(self.gmi, self.group_id, self.message_id)

    def unlike(self):
        LikeDestroyRequest(self.gmi, self.group_id, self.message_id)


class MessagesIndexRequest(Request):
    def __init__(self, gmi, group_id, before_id=None, since_id=None, after_id=None, limit=20):
        self.group_id = group_id
        self.before_id = before_id
        self.since_id = since_id
        self.after_id = after_id
        self.limit = limit
        if limit > 100:
            raise ValueError('Limit must be at or below 100')
        arg_count = 0
        if before_id is not None:
            arg_count += 1
        if since_id is not None:
            arg_count += 1
        if after_id is not None:
            arg_count += 1
        elif arg_count > 1:
            raise ValueError('Only one of before_id, since_id or after_id can be defined')
        super().__init__(gmi)

    def mode(self):
        return "GET"

    def url(self):
        return self.base_url + '/groups/' + str(self.group_id) + '/messages'

    def args(self):
        args_dict = {}
        if self.before_id is not None:
            args_dict['before_id'] = self.before_id
        if self.since_id is not None:
            args_dict['since_id'] = self.since_id
        if self.after_id is not None:
            args_dict['after_id'] = self.after_id
        if self.limit is not 20:
            args_dict['limit'] = self.limit
        return args_dict

    def parse(self, response):
        # count = int(response['count'])
        messages = []
        for message_json in response['messages']:
            messages.append(Message.from_json(self.gmi, message_json))
        return messages


class MessagesCreateRequest(Request):
    def __init__(self, gmi, group_id, source_guid, text, attachments=None):
        self.group_id = group_id
        self.source_guid = source_guid
        self.text = text
        self.attachments = attachments
        super().__init__(gmi)

    def mode(self):
        return "POST"

    def url(self):
        return self.base_url + '/groups/' + str(self.group_id) + '/messages'

    def args(self):
        return {
            'message': {
                'source_guid': self.source_guid,
                'text': self.text,
                'attachments': self.attachments
            }
        }

    def parse(self, response):
        return Message.from_json(self.gmi, response['message'])


# --- Undocumented ---

class MessagesShowRequest(Request):
    def __init__(self, gmi, group_id, message_id):
        self.group_id = group_id
        self.message_id = message_id
        super().__init__(gmi)

    def url(self):
        return self.base_url + '/groups/' + str(self.group_id) + '/messages/' + str(self.message_id)

    def mode(self):
        return "GET"

    def parse(self, response):
        return Message.from_json(self.gmi, response['message'])



from lowerpines.endpoints.object import AbstractObject, Field, RetrievableObject
from lowerpines.endpoints.request import Request
from lowerpines.endpoints.sms import SmsCreateRequest, SmsDeleteRequest


class User(AbstractObject, RetrievableObject):
    user_id = Field()
    phone_number = Field()
    image_url = Field()
    name = Field()
    created_at = Field()
    updated_at = Field()
    email = Field()
    sms = Field()

    def __init__(self, gmi):
        self.gmi = gmi

    def save(self):
        new_data = UserUpdateRequest(self.gmi, self.image_url, self.name, self.email).result
        self._refresh_from_other(new_data)

    def refresh(self):
        new_data = UserMeRequest(self.gmi).result
        self._refresh_from_other(new_data)

    @classmethod
    def get(cls, gmi):
        user = cls(gmi)
        user.refresh()
        return user

    def enable_sms(self, duration, registration_id):
        SmsCreateRequest(self.gmi, duration, registration_id)

    def disable_sms(self):
        SmsDeleteRequest(self.gmi)


class UserMeRequest(Request):
    def mode(self):
        return "GET"

    def parse(self, response):
        return User.from_json(self.gmi, response)

    def url(self):
        return self.base_url + '/users/me'


class UserUpdateRequest(Request):
    def __init__(self, gmi, avatar_url=None, name=None, email=None, zip_code=None):
        self.avatar_url = avatar_url
        self.name = name
        self.email = email
        self.zip_code = zip_code
        super().__init__(gmi)

    def mode(self):
        return "POST"

    def parse(self, response):
        return User.from_json(self.gmi, response)

    def url(self):
        return self.base_url + '/users/update'

    def args(self):
        arg_dict = dict()

        if self.avatar_url:
            arg_dict['avatar_url'] = self.avatar_url
        if self.name:
            arg_dict['name'] = self.name
        if self.email:
            arg_dict['email'] = self.email
        if self.zip_code:
            arg_dict['zip_code'] = self.zip_code

        return arg_dict



from unittest import TestCase

from lowerpines.message import RefAttach, ComplexMessage, smart_split_complex_message, ImageAttach, LocationAttach, \
    SplitAttach, EmojiAttach


class ComplexMessageTest(TestCase):
    def test_manual_creation(self):
        message = ComplexMessage(['Hello, ', RefAttach('user_id_here', '@world')])

        self.assertEqual(message.get_text(), 'Hello, @world')
        self.assertEqual(message.get_attachments(), [{
            'loci': [[7, 6]],
            'type': 'mentions',
            'user_ids': ['user_id_here']
        }])

    def test_dynamic_creation(self):
        message = 'Hello, ' + RefAttach('user_id_here', '@world')

        self.assertEqual(message.get_text(), 'Hello, @world')
        self.assertEqual(message.get_attachments(), [{
            'loci': [[7, 6]],
            'type': 'mentions',
            'user_ids': ['user_id_here']
        }])

    def test_dynamic_creation_reverse_order(self):
        message = RefAttach('user_id_here', '@world') + ' how are you?'

        self.assertEqual(message.get_text(), '@world how are you?')
        self.assertEqual(message.get_attachments(), [{
            'loci': [[0, 6]],
            'type': 'mentions',
            'user_ids': ['user_id_here']
        }])


class SmartSplitComplexMessage(TestCase):
    def test_complex_message(self):
        message = ComplexMessage(['Hello, ', RefAttach('user_id_here', '@world')])
        text, attachments = smart_split_complex_message(message)
        self.assertEqual(text, 'Hello, @world')
        self.assertEqual(attachments, [{
            'loci': [[7, 6]],
            'type': 'mentions',
            'user_ids': ['user_id_here']
        }])

    def test_str_message(self):
        text, attachments = smart_split_complex_message('Hello!')
        self.assertEqual(text, 'Hello!')
        self.assertEqual(attachments, [])


class MessageAttachTest(TestCase):
    def test_mixing_together(self):
        message = RefAttach('user1') + ImageAttach('http://image.url') + LocationAttach('home', 32, 83) + SplitAttach(
            'token') + EmojiAttach(13)
        self.assertEqual(message.get_text(), 'hometoken\u200b')
        self.assertEqual(message.get_attachments(), [
            {
                'type': 'mentions',
                'loci': [[0, 0]],
                'user_ids': ['user1'],
            },
            {
                'type': 'image',
                'url': 'http://image.url',
            },
            {
                'type': 'location',
                'name': 'home',
                'lat': 32,
                'long': 83,
            },
            {
                'type': 'split',
                'token': 'token',
            },
            {
                'type': 'emoji',
                'charmap': [[13, 9]],
                'placeholder': '\u200b',
            }
        ])


class RefAttachTest(TestCase):
    def test_hidden(self):
        message = 'Test' + RefAttach('user_id_here')
        self.assertEqual(message.get_text(), 'Test')
        self.assertEqual(message.get_attachments(), [{
            'loci': [[4, 0]],
            'type': 'mentions',
            'user_ids': ['user_id_here']
        }])

    def test_visible(self):
        message = 'Test ' + RefAttach('user_id_here', '@all')
        self.assertEqual(message.get_text(), 'Test @all')
        self.assertEqual(message.get_attachments(), [{
            'loci': [[5, 4]],
            'type': 'mentions',
            'user_ids': ['user_id_here']
        }])

    def test_multiple_in_order(self):
        message = 'Test ' + RefAttach('user1', '@1') + RefAttach('user2', '@2') + RefAttach('user3', '@3')
        self.assertEqual(message.get_text(), 'Test @1@2@3')
        self.assertEqual(message.get_attachments(), [{
            'loci': [[5, 2], [7, 2], [9, 2]],
            'type': 'mentions',
            'user_ids': ['user1', 'user2', 'user3']
        }])

    def test_multiple_split(self):
        message = RefAttach('red_id', '@red') + ' vs. ' + RefAttach('blue_id', '@blue')
        self.assertEqual(message.get_text(), '@red vs. @blue')
        self.assertEqual(message.get_attachments(), [{
            'type': 'mentions',
            'loci': [[0, 4], [9, 5]],
            'user_ids': ['red_id', 'blue_id']
        }])


class ImageAttachTest(TestCase):
    def test_order_independence(self):
        message1 = ImageAttach('image_url_here') + 'Check out my cool image!'
        message2 = 'Check out my cool image!' + ImageAttach('image_url_here')

        self.assertEqual(message1.get_text(), message2.get_text())
        self.assertEqual(message1.get_attachments(), message2.get_attachments())



from distutils.core import setup

import lowerpines

setup(
    name='lowerpines',
    packages=['lowerpines', 'lowerpines.endpoints'],
    version=lowerpines.VERSION,
    description='Library wrapper for GroupMe API',
    requires=['requests'],
    license='GNU Lesser General Public License v3 (LGPLv3)',
    author='Jonathan Janzen',
    author_email='jjjonjanzen@gmail.com',
    url='https://github.com/bigfootjon/lowerpines',
    download_url='https://github.com/bigfootjon/lowerpines/tarball/' + lowerpines.VERSION,
    keywords=['api', 'GroupMe'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
        'Topic :: Communications :: Chat',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
)



from datetime import datetime

from lowerpines.endpoints.object import AbstractObject, Field, RetrievableObject
from lowerpines.endpoints.request import Request
from lowerpines.endpoints.member import MembersAddRequest, MembersRemoveRequest, Member
from lowerpines.endpoints.message import Message
from lowerpines.exceptions import InvalidOperationException
from lowerpines.message import smart_split_complex_message


class Group(AbstractObject, RetrievableObject):
    group_id = Field(api_name='id')
    name = Field()
    type = Field()
    description = Field()
    image_url = Field()
    creator_user_id = Field()
    created_at = Field()
    updated_at = Field()
    share_url = Field()
    members_raw = Field(api_name='members', handler=None)
    messages_count_raw = Field(api_name='messages.count', handler=None)
    messages_last_message_id_raw = Field(api_name='messages.last_message_id', handler=None)
    messages_last_message_created_at_raw = Field(api_name='messages.last_message_created_at', handler=None)

    def __init__(self, gmi, name=None, description=None, image_url=None):
        super().__init__()
        self.gmi = gmi
        self.messages = GroupMessagesManager(self)
        self.name = name
        self.description = description
        self.image_url = image_url
        self.members = []

    @property
    def bots(self):
        return self.gmi.bots.filter(group_id=self.group_id)

    def on_fields_loaded(self):
        self.members = []
        for member_json in self.members_raw:
            self.members.append(Member.from_json(self.gmi, member_json, self.group_id))
        self.messages.count = self.messages_count_raw
        self.messages.last_id = self.messages_last_message_id_raw
        self.messages.last_created_at = self.messages_last_message_created_at_raw

    def save(self):
        if self.group_id is None:
            new_data = GroupsCreateRequest(self.gmi, self.name, self.description, self.image_url).result
        else:
            new_data = GroupsUpdateRequest(self.gmi, self.group_id, self.name, self.description, self.image_url).result

        self._refresh_from_other(new_data)

    def delete(self):
        if self.group_id is None:
            raise InvalidOperationException('Cannot destroy a group that isn\'t saved!')
        else:
            GroupsDestroyRequest(self.gmi, self.group_id)

    def refresh(self):
        if self.group_id is None:
            raise InvalidOperationException('Must have an id to perform this operation')
        else:
            new_data = GroupsShowRequest(self.gmi, group_id=self.group_id).result
            self._refresh_from_other(new_data)

    def member_add(self, name, user_id):
        MembersAddRequest(self.gmi, self.group_id, name, user_id=user_id)

    def member_rm(self, member_id):
        MembersRemoveRequest(self.gmi, self.group_id, member_id)

    def post(self, message):
        text, attachments = smart_split_complex_message(message)
        obj = Message(self.gmi, self.group_id, str(datetime.now()), text, attachments)
        obj.save()
        return obj

    @staticmethod
    def get(gmi, group_id):
        return GroupsShowRequest(gmi, group_id).result

    @staticmethod
    def get_all(gmi):
        return GroupsIndexRequest(gmi, per_page=100).result

    @staticmethod
    def get_former(gmi):
        return GroupsFormerRequest(gmi).result

    @staticmethod
    def join(gmi, group_id, share_token):
        return GroupsJoinRequest(gmi, group_id, share_token).result

    @staticmethod
    def rejoin(gmi, group_id):
        return GroupsRejoinRequest(gmi, group_id).result

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)


class GroupMessagesManager:
    count = 0

    last_id = str()
    last_created_at = str()

    def __init__(self, group):
        self.group = group

    def all(self):
        messages = self.recent()
        while len(messages) < self.count:
            messages.extend(self.before(messages[-1]))
        return messages

    def recent(self, count=100):
        from lowerpines.endpoints.message import MessagesIndexRequest
        return MessagesIndexRequest(self.group.gmi, self.group.group_id, limit=count).result

    def before(self, message, count=100):
        from lowerpines.endpoints.message import MessagesIndexRequest
        return MessagesIndexRequest(self.group.gmi, self.group.group_id, limit=count,
                                    before_id=message.message_id).result

    def since(self, message, count=100):
        from lowerpines.endpoints.message import MessagesIndexRequest
        return MessagesIndexRequest(self.group.gmi, self.group.group_id, limit=count,
                                    since_id=message.message_id).result

    def after(self, message, count=100):
        from lowerpines.endpoints.message import MessagesIndexRequest
        return MessagesIndexRequest(self.group.gmi, self.group.group_id, limit=count,
                                    after_id=message.message_id).result


class GroupsIndexRequest(Request):
    def __init__(self, gmi, page=1, per_page=10):
        self.page = page
        self.per_page = per_page
        super().__init__(gmi)

    def url(self):
        return self.base_url + "/groups"

    def args(self):
        return {
            'page': self.page,
            'per_page': self.per_page
        }

    def mode(self):
        return "GET"

    def parse(self, response):
        groups = []
        for group_json in response:
            groups.append(Group.from_json(self.gmi, group_json))
        return groups


class GroupsFormerRequest(Request):
    def url(self):
        return self.base_url + "/groups/former"

    def args(self):
        return {}

    def mode(self):
        return "GET"

    def parse(self, response):
        groups = []
        for group_json in response:
            groups.append(Group.from_json(self.gmi, group_json))
        return groups


class GroupsShowRequest(Request):
    def __init__(self, gmi, group_id):
        self.group_id = group_id
        super().__init__(gmi)

    def url(self):
        return self.base_url + "/groups/" + str(self.group_id)

    def args(self):
        return {}

    def mode(self):
        return "GET"

    def parse(self, response):
        return Group.from_json(self.gmi, response)


class GroupsCreateRequest(Request):
    def parse(self, response):
        return Group.from_json(self.gmi, response)

    def __init__(self, gmi, name, description=None, image_url=None, share=None):
        self.name = name
        self.description = description
        self.image_url = image_url
        self.share = share
        super().__init__(gmi)

    def url(self):
        return self.base_url + "/groups"

    def mode(self):
        return "POST"

    def args(self):
        post_args = {
            'name': self.name
        }
        if self.description is not None:
            post_args['description'] = self.description
        if self.image_url is not None:
            post_args['image_url'] = self.image_url
        if self.share is not None:
            post_args['share'] = self.share
        return post_args


class GroupsUpdateRequest(Request):
    def __init__(self, gmi, group_id, name=None, description=None, image_url=None, office_mode=None, share=None):
        self.group_id = group_id
        self.name = name
        self.description = description
        self.image_url = image_url
        self.office_mode = office_mode
        self.share = share
        super().__init__(gmi)

    def parse(self, response):
        return Group.from_json(self.gmi, response)

    def url(self):
        return self.base_url + "/groups/" + str(self.group_id) + '/update'

    def mode(self):
        return "POST"

    def args(self):
        arg_dict = {}
        if self.name is not None:
            arg_dict['name'] = self.name
        if self.description is not None:
            arg_dict['description'] = self.description
        if self.image_url is not None:
            arg_dict['image_url'] = self.image_url
        if self.office_mode is not None:
            arg_dict['office_mode'] = self.office_mode
        if self.share is not None:
            arg_dict['share'] = self.share
        return arg_dict


class GroupsDestroyRequest(Request):
    def __init__(self, gmi, group_id):
        self.group_id = group_id
        super().__init__(gmi)

    def parse(self, response):
        pass

    def url(self):
        return self.base_url + '/groups/' + str(self.group_id) + '/destroy'

    def mode(self):
        return 'POST'


class GroupsJoinRequest(Request):
    def __init__(self, gmi, group_id, share_token):
        self.group_id = group_id
        self.share_token = share_token
        super().__init__(gmi)

    def parse(self, response):
        return Group.from_json(self.gmi, response)

    def url(self):
        return self.base_url + '/groups/' + str(self.group_id) + '/join/' + str(self.share_token)

    def mode(self):
        return "POST"


class GroupsRejoinRequest(Request):
    def __init__(self, gmi, group_id):
        self.group_id = group_id
        super().__init__(gmi)

    def parse(self, response):
        return Group.from_json(self.gmi, response)

    def url(self):
        return self.base_url + '/groups/join'

    def mode(self):
        return "POST"

    def args(self):
        return {
            'group_id': self.group_id
        }



from lowerpines.endpoints.message import Message


class MessageManager:
    def __init__(self, gmi):
        self.gmi = gmi

    def get(self, group_id, message_id):
        return Message.get(self.gmi, group_id, message_id)


class MessageAttach:
    def __add__(self, other):
        if isinstance(other, ComplexMessage):
            other.contents.append(self)
            return other
        else:
            return ComplexMessage([self, other])

    def __radd__(self, other):
        if isinstance(other, ComplexMessage):
            other.contents.append(self)
            return other
        else:
            return ComplexMessage([other, self])


class RefAttach(MessageAttach):
    def __init__(self, user_id, display=''):
        self.display = display
        self.user_id = user_id

    def __str__(self):
        return self.display

    def __repr__(self):
        return 'R:' + self.display


class ImageAttach(MessageAttach):
    def __init__(self, image_url):
        self.image_url = image_url

    def __str__(self):
        return ''

    def __repr__(self):
        return 'I:' + str(self)


class LocationAttach(MessageAttach):
    def __init__(self, name, lat, long):
        self.name = name
        self.lat = lat
        self.long = long

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'L:' + str(self)


# This feature is not supported anymore but it still exists in the API
class SplitAttach(MessageAttach):
    def __init__(self, token):
        self.token = token

    def __str__(self):
        return self.token

    def __repr__(self):
        return 'S:' + str(self)


EMOJI_PLACEHOLDER = '\u200B'


class EmojiAttach(MessageAttach):
    def __init__(self, pack_id):
        self.pack_id = pack_id

    def __str__(self):
        return EMOJI_PLACEHOLDER

    def __repr__(self):
        return 'E:' + str(self)


class ComplexMessage:
    def __init__(self, data):
        if isinstance(data, list):
            self.contents = data
        else:
            self.contents = [data]

    def __add__(self, other):
        if isinstance(other, self.__class__):
            self.contents.extend(other.contents)
        else:
            self.contents.append(other)
        return self

    def __radd__(self, other):
        if isinstance(other, self.__class__):
            self.contents = other.contents + self.contents
        else:
            self.contents.insert(0, other)
        return self

    def __str__(self):
        return str(self.contents)

    def get_text(self):
        return ''.join([str(part) for part in self.contents])

    def get_attachments(self):
        attach_list = []
        mentions = {
            'type': 'mentions',
            'user_ids': list(),
            'loci': list()
        }
        emojis = {
            'type': 'emoji',
            'placeholder': EMOJI_PLACEHOLDER,
            'charmap': []
        }
        content_frag = ""
        for part in self.contents:
            if isinstance(part, RefAttach):
                mentions['user_ids'].append(part.user_id)
                mentions['loci'].append([len(content_frag), len(part.display)])
                if mentions not in attach_list:
                    attach_list.append(mentions)
            elif isinstance(part, ImageAttach):
                attach_list.append({
                    'type': 'image',
                    'url': part.image_url
                })
            elif isinstance(part, LocationAttach):
                attach_list.append({
                    'type': 'location',
                    'name': part.name,
                    'lat': part.lat,
                    'long': part.long
                })
            elif isinstance(part, SplitAttach):
                attach_list.append({
                    'type': 'split',
                    'token': part.token
                })
            elif isinstance(part, EmojiAttach):
                emojis['charmap'].append([part.pack_id, len(content_frag)])
                if emojis not in attach_list:
                    attach_list.append(emojis)
            content_frag += str(part)
        return attach_list

    def just_str(self):
        return ''.join([s for s in self.contents if isinstance(s, str)])


def smart_split_complex_message(message):
    if isinstance(message, ComplexMessage):
        return message.get_text(), message.get_attachments()
    elif isinstance(message, str):
        return message, []
    else:
        raise ValueError('Message object must be a str or ComplexMessage')
